--[[
-- TODO LIST:
--1. Option to add custom textures for border
--4. Option to lock a panel to the visible frame area
--5. Option to copy settings from one panel to another
--6. Option to change offset of gradient
--7. Option to hide a frame (decide if i really want it)
--8. Ability to adjust frame size by cursor when unlocked
--9. Look into hooking a panels anchor to another visible GUI element via mouse
--   (does visor do this?  just let ppl do it using visor?)
--]]





--[[
-- Confirm + reset all panels
--]]
StaticPopupDialogs["EEPANELSRESET"] = {
  text = "Reset?",
  button1 = "Yes",
  button2 = "No",
  OnAccept = function() eePanels:Reset() end,
  timeout = 0,
  whileDead = 1,
  hideOnEscape = 1
}


--[[
-- Round a number to the nearest integer
--]]
function math.round( n )
   dec_places = 0
	shift = 10^dec_places
	return math.floor ((n*shift)+0.5)/shift
end


--[[function deepCopyInto(from_table, copy_table, lookup_table)
   for i,v in pairs(from_table) do
      --if i == "frame" or i == "guideFrame" then continue end
      if type(v) ~= "table" and i ~= "frame" and i ~= "guideFrame" then
         copy_table[i] = v
      else
         lookup_table = lookup_table or {}
         lookup_table[from_table] = copy_table
         if lookup_table[v] then
            copy_table[i] = lookup_table[v]
         else
            deepCopyInto(v, copy_table[i], lookup_table)
         end
      end
   end
end--]]

--[[
-- Return a deep copy of a table
--]]
function deepCopy(t, lookup_table)
   local copy = {}
   for i,v in pairs(t) do
      if type(v) ~= "table" then
         copy[i] = v
      else
         lookup_table = lookup_table or {}
         lookup_table[t] = copy
         if lookup_table[v] then
            copy[i] = lookup_table[v] -- we already copied this table. reuse the copy.
         else
            copy[i] = deepCopy(v,lookup_table) -- not yet copied. copy it.
         end
      end
   end
   return copy
end


--[[
-- Returns the key associated with val in table
--]]
function rLookup(table,val)
   for k,v in pairs(table) do
      if v == val then return k end
   end
   return nil
end




-- User selectable options
local bgColorStyle = { solid="Solid", gradient="Gradient" }
local bgGradOrientation = { horizontal="HORIZONTAL", vertical="VERTICAL"}
local frameStrata = { background="BACKGROUND", low="LOW", medium="MEDIUM", high="HIGH", dialog="DIALOG", tooltip="TOOLTIP" }
local blendMode = { disable="DISABLE", blend="BLEND", add="ADD", mod="MOD" }
local borderTexture = { none = "", tooltip = "Interface\\Tooltips\\UI-Tooltip-Border", dialog = "Interface\\DialogFrame\\UI-DialogBox-Border" }


-- Generate validation tables
local val_bgColorStyle = {}
for k,v in pairs(bgColorStyle) do table.insert(val_bgColorStyle,k) end
local val_bgGradOrientation ={}
for k,v in pairs(bgGradOrientation) do table.insert(val_bgGradOrientation,k) end
local val_frameStrata = {}
for k,v in pairs(frameStrata) do table.insert(val_frameStrata,k) end
local val_blendMode = {}
for k,v in pairs(blendMode) do table.insert(val_blendMode,k) end
local val_borderTexture = {}
for k,v in pairs(borderTexture) do table.insert(val_borderTexture,k) end


-- LEAVING THIS IN FOR NOW FOR OLD PROFILES WHICH REFERENCE IT
local backgroundTexture = { none = "", tooltip = "Interface\\Tooltips\\UI-Tooltip-Background", dialog = "Interface\\DialogFrame\\UI-DialogBox-Background" }
local val_backgroundTexture = {}
for k,v in pairs(backgroundTexture) do table.insert(val_backgroundTexture,k) end
      
   
-- Default panel settings
local dp = 
{
   frame = nil,
   guideFrame = nil,
   name = "",
   x = 0,
   y = 0,
   width = 200,
   height = 100,
   visible = true,
   level = 0,
   strata = frameStrata.background,
   --texture = backgroundTexture.none,
   texture = "",
   textureAlpha = 1,
   border = 
   {
      color = {r=1,g=1,b=1,a=1},
      texture = borderTexture.tooltip
   },
   background = 
   {
      frame = nil,
      style = bgColorStyle.solid,
      color = {r=0,g=0,b=0,a=0.8},
      gradient =
      {
         blend = blendMode.disable, -- should have been in table above, not in gradient
         offset = 0,
         color = {r=1,g=1,b=1,a=0},
         orientation = bgGradOrientation.horizontal
      }
   },
}


--[[
-- AceDB table
--]]
local defaults = 
{
   panels = {},
   defaultPanel = deepCopy(dp),
   isLocked = true,
}






--[[
-- Initialize our addon
--]]
eePanels = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceConsole-2.0", "AceDB-2.0", "FuBarPlugin-2.0")


--[[
-- AceOption table
--]]
eePanels.menu = 
{ 
   type='group',
   args = {
      
      lock = {
         type = 'toggle',
         name = "Lock",
         desc = "LockDesc",
         get  = function() return eePanels.db.profile.isLocked end,
         set  = function(v) eePanels:ToggleLock() end,
         order = 1,
      },
      
      createDefaultPanel = {
         type = 'execute',
         name = 'Create Default Panel',
         desc = 'Create a panel using current default settings',
         func = function(v) eePanels:CreatePanel(1) end,
         order = 2
      },
      
      createEmptyPanel = {
         type = 'execute',
         name = 'Create Empty Panel',
         desc = 'Create a panel using empty settings',
         func = function(v) eePanels:CreatePanel(2) end,
         order = 3
      },
      
      allPanels = {
         type = 'group',
         name = 'All Panels',
         desc = 'Change default panel settings, and settings for any current panels',
         func = function() end,
         hidden = function() return not (table.getn(eePanels.db.profile.panels) > 0) end,
         order = 4,
         args = {
            
            borderColor = {
               name = 'Border Color', desc = 'Border Color', type='color',
               get = function()
                  local panel = eePanels.db.profile.defaultPanel
                  return panel.border.color.r,panel.border.color.g,panel.border.color.b,panel.border.color.a 
               end,
               set = function(ir,ig,ib,ia)
                  local panel = eePanels.db.profile.defaultPanel
                  panel.border.color = {r=ir,g=ig,b=ib,a=ia} 
                  for i in pairs(eePanels.db.profile.panels) do
                     eePanels.db.profile.panels[i].border.color = {r=ir,g=ig,b=ib,a=ia} 
                     eePanels:ChangeBorderColor(eePanels.db.profile.panels[i]) 
                  end
               end,
               hasAlpha = true,
               order = 10,
            },
            
            borderTexture = {
               name = 'Border Texture', desc = 'Border Texture', type='text',
               get = function() 
                  local panel = eePanels.db.profile.defaultPanel
                  return rLookup(borderTexture,panel.border.texture) 
               end,
               set = function(t) 
                  local panel = eePanels.db.profile.defaultPanel
                  panel.border.texture = borderTexture[t] 
                  for i in pairs(eePanels.db.profile.panels) do
                     eePanels.db.profile.panels[i].border.texture = borderTexture[t] 
                     eePanels:ChangeBackdrop(eePanels.db.profile.panels[i]) 
                     eePanels:ChangeBorderColor(eePanels.db.profile.panels[i]) 
                  end
               end,
               validate = val_borderTexture,
               order = 10,
            },
            
            backgroundColorStyle = {
               name = 'Background Color Style', desc = 'Background Color Style', type='text',
               get = function() 
                  local panel = eePanels.db.profile.defaultPanel
                  return panel.background.style 
               end,
               set = function(t) 
                  local panel = eePanels.db.profile.defaultPanel
                  panel.background.style = bgColorStyle[t] 
                  for i in pairs(eePanels.db.profile.panels) do
                     eePanels.db.profile.panels[i].background.style = bgColorStyle[t]
                     eePanels:ChangeBackdrop(eePanels.db.profile.panels[i]) 
                     eePanels:ChangeBackgroundColor(eePanels.db.profile.panels[i]) 
                     eePanels:ChangeBorderColor(eePanels.db.profile.panels[i]) 
                  end
               end,
               validate = val_bgColorStyle,
               order = 11,
            },
            
            backgroundColor = {
               name = 'Background Color', desc = 'Background Color', type='color',
               get = function() 
                  local panel = eePanels.db.profile.defaultPanel
                  return panel.background.color.r,panel.background.color.g,panel.background.color.b,panel.background.color.a 
               end,
               set = function(ir,ig,ib,ia)
                  local panel = eePanels.db.profile.defaultPanel
                  panel.background.color = {r=ir,g=ig,b=ib,a=ia} 
                  for i in pairs(eePanels.db.profile.panels) do
                     eePanels.db.profile.panels[i].background.color = {r=ir,g=ig,b=ib,a=ia} 
                     eePanels:ChangeBackgroundColor(eePanels.db.profile.panels[i]) 
                  end
               end,
               hasAlpha = true,
               order = 12,
            },
            
            backgroundGradientColor = {
               name = 'Background Gradient Color', desc = 'Background Gradient Color', type='color',
               get = function() 
                  local panel = eePanels.db.profile.defaultPanel
                  return panel.background.gradient.color.r,panel.background.gradient.color.g,panel.background.gradient.color.b,panel.background.gradient.color.a 
               end,
               set = function(ir,ig,ib,ia)
                  local panel = eePanels.db.profile.defaultPanel
                  panel.background.gradient.color = {r=ir,g=ig,b=ib,a=ia}
                  for i in pairs(eePanels.db.profile.panels) do
                     eePanels.db.profile.panels[i].background.gradient.color = {r=ir,g=ig,b=ib,a=ia} 
                     eePanels:ChangeBackgroundColor(eePanels.db.profile.panels[i])
                  end
               end,
               hasAlpha = true,
               order = 13,
            },
            
            backgroundGradientOrientation = {
               name = 'Background Gradient Orientation', desc = 'Background Gradient Orientation', type='text',
               get = function() 
                  local panel = eePanels.db.profile.defaultPanel
                  return panel.background.gradient.orientation 
               end,
               set = function(t) 
                  local panel = eePanels.db.profile.defaultPanel
                  panel.background.gradient.orientation = bgGradOrientation[t]
                  for i in pairs(eePanels.db.profile.panels) do
                     eePanels.db.profile.panels[i].background.gradient.orientation = bgGradOrientation[t]
                     eePanels:ChangeBackgroundColor(eePanels.db.profile.panels[i]) 
                  end
               end,
               validate = val_bgGradOrientation,
               order = 14,
            },
            
            backgroundTexture = {
               name = 'Background Texture', desc = 'Background Texture', type='text',usage='Path to texture',
               get = function() 
                  local panel = eePanels.db.profile.defaultPanel
                  return panel.texture
               end,
               set = function(t) 
                  local panel = eePanels.db.profile.defaultPanel
                  panel.texture = t
                  for i in pairs(eePanels.db.profile.panels) do
                     eePanels.db.profile.panels[i].texture = t
                     eePanels:ChangeBackdrop(eePanels.db.profile.panels[i]) 
                     eePanels:ChangeBorderColor(eePanels.db.profile.panels[i]) 
                  end
               end,
               --set = function(t) panel.texture = t eePanels:ChangeBackdrop(panel) eePanels:ChangeBorderColor(panel) end,
               order = 15,
            },
            
            backgroundBlend = {
               name = 'Background Blend', desc = 'Background Blend', type='text',
               get = function()
                  local panel = eePanels.db.profile.defaultPanel
                  return panel.background.gradient.blend
               end,
               set = function(b)
                  local panel = eePanels.db.profile.defaultPanel
                  panel.background.gradient.blend = blendMode[b]
                  for i in pairs(eePanels.db.profile.panels) do
                     eePanels.db.profile.panels[i].background.gradient.blend = blendMode[b]
                     eePanels:ChangeTextureBlend(eePanels.db.profile.panels[i])
                  end
               end,
               --get = function() return panel.background.gradient.blend end,
               --set = function(b) panel.background.gradient.blend = blendMode[b] eePanels:ChangeTextureBlend(panel) end,
               validate = val_blendMode,
               order = 16,
            },
            
            level = {
               name = 'Level', desc = 'Panel Level', type='range',
               get = function()
                  local panel = eePanels.db.profile.defaultPanel
                  if not panel.level then panel.level = 0 end
                  return panel.level
               end,
               set = function(l)
                  local panel = eePanels.db.profile.defaultPanel
                  panel.level = l
                  for i in pairs(eePanels.db.profile.panels) do
                     eePanels.db.profile.panels[i].level = l
                     eePanels:ChangeLevel(eePanels.db.profile.panels[i])
                  end
               end,
               min = 0, max = 20, step = 1, isPercent = false,
               order = 17,
            },
            
            strata = {
               name = 'Strata', desc = 'Panel Level', type='text', 
               get = function() 
                  local panel = eePanels.db.profile.defaultPanel
                  return panel.strata 
               end,
               set = function(u) 
                  local panel = eePanels.db.profile.defaultPanel
                  panel.strata = frameStrata[u] 
                  for i in pairs(eePanels.db.profile.panels) do
                     eePanels.db.profile.panels[i].strata = frameStrata[u] 
                     eePanels:ChangeStrata(eePanels.db.profile.panels[i]) 
                  end
               end,
               validate = val_frameStrata,
               order = 18,
            },
            
            width = {
               name = 'Width', desc = 'Width', type='text', usage=' ',
               get = function() 
                  local panel = eePanels.db.profile.defaultPanel
                  return math.round(panel.width) 
               end,
               set = function(u) 
                  local panel = eePanels.db.profile.defaultPanel
                  panel.width = u 
                  for i in pairs(eePanels.db.profile.panels) do
                     eePanels.db.profile.panels[i].width = u 
                     eePanels:ChangeWidth(eePanels.db.profile.panels[i])
                  end
               end,
               order = 19,
            },
            
            height = {
               name = 'Height', desc = 'Height', type='text', usage=' ',
               get = function() 
                  local panel = eePanels.db.profile.defaultPanel
                  return math.round(panel.height) 
               end,
               set = function(u) 
                  local panel = eePanels.db.profile.defaultPanel
                  panel.height = u 
                  for i in pairs(eePanels.db.profile.panels) do
                     eePanels.db.profile.panels[i].height = u 
                     eePanels:ChangeHeight(eePanels.db.profile.panels[i]) 
                  end
               end,
               order = 20,
            },
            
            x = {
               name = 'Horizontal Position', desc = 'Horizontal Position', type='text', usage=' ',
               get = function() 
                  local panel = eePanels.db.profile.defaultPanel
                  return math.round(panel.x) 
               end,
               set = function(u) 
                  local panel = eePanels.db.profile.defaultPanel
                  panel.x = u 
                  for i in pairs(eePanels.db.profile.panels) do
                     eePanels.db.profile.panels[i].x = u 
                     eePanels:ChangePosition(eePanels.db.profile.panels[i]) 
                  end
               end,
               order = 21,
            },
            
            y = {
               name = 'Vertical Position', desc = 'Vertical Position', type='text', usage=' ',
               get = function() 
                  local panel = eePanels.db.profile.defaultPanel
                  return math.round(panel.y) 
               end,
               set = function(u) 
                  local panel = eePanels.db.profile.defaultPanel
                  panel.y = u 
                  for i in pairs(eePanels.db.profile.panels) do
                     eePanels.db.profile.panels[i].y = u 
                     eePanels:ChangePosition(eePanels.db.profile.panels[i]) 
                  end
               end,
               order = 22,
            },
            
         }
      },
      
      panels = {
         type = 'group',
         name = 'Panels',
         desc = 'Panels',
         hidden = function() return not (table.getn(eePanels.db.profile.panels) > 0) end,
         order = 5,
         args = {}
      },
      
      space1 = {
         name = " ",
         type = 'header',
         order = 6,
      },
         
      reset = {
         type = 'execute',
         name = 'Reset',
         desc = 'Reset',
         func = function(v) StaticPopup_Show("EEPANELSRESET") end,
         order = 7,
      },
      
   }
}


eePanels.OnMenuRequest = eePanels.menu
eePanels.name = "eePanels"
eePanels.hasIcon = true
eePanels.defaultMinimapPosition = 35
eePanels.cannotDetachTooltip = true
eePanels.independentProfile = true
eePanels.defaultPosition = "LEFT"
eePanels.hideWithoutStandby = true
eePanels.hasNoColor = true


--[[
-- Initialize addon settings for ace2
--]]
function eePanels:OnInitialize()
   self:RegisterDB("eePanelsDB")
	self:RegisterDefaults('profile', defaults )
	self:RegisterChatCommand( {"/eePanels"}, eePanels.menu )
	if not FuBar then
		self.OnMenuRequest.args.hide.guiName = "Hide minimap icon"
		self.OnMenuRequest.args.hide.desc = "Hide minimap icon"
	end
end


--[[
-- Restore each panel in the database 
-- Display the guide frame if the panel isn't locked
--]]
function eePanels:OnEnable()
   for i in pairs(eePanels.db.profile.panels) do
      eePanels:RestorePanel(eePanels.db.profile.panels[i])
      -- If we're in locked mode, create the guideFrame for this panel
      if not eePanels.db.profile.isLocked then 
         eePanels:CreateGuideFrame(eePanels.db.profile.panels[i],i) 
      end
   end
end


--[[
-- Hide all frames when disabled
--]]
function eePanels:OnDisable()
   for i in pairs(eePanels.db.profile.panels) do
      eePanels.db.profile.panels[i].frame:Hide()
   end
end






--[[
-- Removes a created panel
--]]
function eePanels:RemovePanel(index)
   -- Hide the panel and remove it from out database
   eePanels.menu.args.panels.args["panel_"..index] = {}
   eePanels.db.profile.panels[index].frame:Hide()
   eePanels.db.profile.panels[index] = {}
   table.remove(eePanels.db.profile.panels,index)
   
   -- Recreate individual panel option tables
   eePanels:CreateMenus()
   
   -- Toggle lock; forces guidePanels to re-generate the name displayed on the guideFrame
   if not eePanels.db.profile.isLocked then
      eePanels:ToggleLock()--lock
      eePanels:ToggleLock()--unlock
   end
end


--[[
-- Create a new panel
--]]
function eePanels:CreatePanel(defType)
   -- Create a new panel from the default template in the database
   --local newPanel = deepCopy(eePanels.db.profile.defaultPanel)
   local newPanel;
   if not defType or defType == 1 then
      newPanel = deepCopy(eePanels.db.profile.defaultPanel)
   else
      newPanel = deepCopy(dp)
   end
   table.insert(eePanels.db.profile.panels, newPanel)
   
   eePanels:CreateFrame(newPanel)
   eePanels:CreateTexture(newPanel)
   eePanels:ChangeStrata(newPanel)
   eePanels:ChangeLevel(newPanel)
   eePanels:ChangeWidth(newPanel)
   eePanels:ChangeHeight(newPanel)
   -- Center the panel, and store it's position
   newPanel.frame:SetPoint("CENTER", UIParent, "CENTER")
   newPanel.x = newPanel.frame:GetLeft()
   newPanel.y = newPanel.frame:GetTop()
   -- Now change it's anchoring to 'normal', keeping its current position
   eePanels:ChangePosition(newPanel)
   eePanels:ChangeBackdrop(newPanel)
   eePanels:ChangeBorderColor(newPanel)
   eePanels:ChangeBackgroundColor(newPanel)
   
    -- If we're in locked mode, create the guideFrame for this panel
   local newi = table.getn(eePanels.db.profile.panels)
   if not eePanels.db.profile.isLocked then 
      eePanels:CreateGuideFrame(eePanels.db.profile.panels[newi],newi) 
   end
   
   eePanels:CreateMenus()
end


--[[
-- Recreates a panel from the database
--]]
function eePanels:RestorePanel(panel)
   eePanels:CreateFrame(panel)
   eePanels:CreateTexture(panel)
   eePanels:ChangeStrata(panel)
   eePanels:ChangeLevel(panel)
   eePanels:ChangeWidth(panel)
   eePanels:ChangeHeight(panel)
   eePanels:ChangePosition(panel)
   eePanels:ChangeBackdrop(panel)
   eePanels:ChangeBorderColor(panel)
   eePanels:ChangeBackgroundColor(panel)
   
   eePanels:CreateMenus()
end


--[[
-- Creates option table for a panel
--]]
function eePanels:CreateMenus()
   -- Empty the current option table for panels
   eePanels.menu.args.panels.args = {}

   for i in pairs(eePanels.db.profile.panels) do
      -- We must explicitly create a variable for any values we need to dynamically pass through our menu functions
      local index = i
      local panel = eePanels.db.profile.panels[i]
      
      -- NEW VARIABLES:
      -- Here we can make sure any new variables are initialized so we don't break existing layouts
      -- We also make sure that the variables are set in any eePanels:ChangeXXX functions
      if not panel.name then panel.name = "" end
      if not panel.level then panel.level = 0 end
      
      -- Create panels option table
      eePanels.menu.args.panels.args["panel_"..i] = 
      {
         name = i..'. '..panel.name, desc = i..'. '..panel.name, type = 'group',
         args = {
            
            panelName = {
               name = 'Name', desc = 'Panel Name', type='text', usage='name',
               get = function() return panel.name end,
               set = function(n) panel.name = n eePanels:CreateMenus() eePanels:ChangeName(panel,i) end,
               order = 9,
            },
            
            borderColor = {
               name = 'Border Color', desc = 'Border Color', type='color',
               get = function() return panel.border.color.r,panel.border.color.g,panel.border.color.b,panel.border.color.a end,
               set = function(ir,ig,ib,ia) panel.border.color = {r=ir,g=ig,b=ib,a=ia} eePanels:ChangeBorderColor(panel) end,
               hasAlpha = true,
               order = 10,
            },
            
            borderTexture = {
               name = 'Border Texture', desc = 'Border Texture', type='text',
               get = function() return rLookup(borderTexture,panel.border.texture) end,
               set = function(t) panel.border.texture = borderTexture[t] eePanels:ChangeBackdrop(panel) eePanels:ChangeBorderColor(panel) end,
               validate = val_borderTexture,
               order = 10,
            },
            
            backgroundColorStyle = {
               name = 'Background Color Style', desc = 'Background Color Style', type='text',
               get = function() return panel.background.style end,
               set = function(t) panel.background.style = bgColorStyle[t] eePanels:ChangeBackdrop(panel) eePanels:ChangeBackgroundColor(panel) eePanels:ChangeBorderColor(panel) end,
               validate = val_bgColorStyle,
               order = 11,
            },
            
            backgroundColor = {
               name = 'Background Color', desc = 'Background Color', type='color',
               get = function() return panel.background.color.r,panel.background.color.g,panel.background.color.b,panel.background.color.a end,
               set = function(ir,ig,ib,ia) panel.background.color = {r=ir,g=ig,b=ib,a=ia} eePanels:ChangeBackgroundColor(panel) end,
               hasAlpha = true,
               order = 12,
            },
            
            backgroundGradientColor = {
               name = 'Background Gradient Color', desc = 'Background Gradient Color', type='color',
               get = function() return panel.background.gradient.color.r,panel.background.gradient.color.g,panel.background.gradient.color.b,panel.background.gradient.color.a end,
               set = function(ir,ig,ib,ia) panel.background.gradient.color = {r=ir,g=ig,b=ib,a=ia} eePanels:ChangeBackgroundColor(panel) end,
               hasAlpha = true,
               order = 13,
            },
            
            backgroundGradientOrientation = {
               name = 'Background Gradient Orientation', desc = 'Background Gradient Orientation', type='text',
               get = function() return panel.background.gradient.orientation end,
               set = function(t) panel.background.gradient.orientation = bgGradOrientation[t] eePanels:ChangeBackgroundColor(panel) end,
               validate = val_bgGradOrientation,
               order = 14,
            },
            
            backgroundTexture = {
               name = 'Background Texture', desc = 'Background Texture', type='text',usage='Path to texture',
               --get = function() return rLookup(backgroundTexture,panel.texture) end,
               get = function() return panel.texture end,
               --set = function(t) panel.texture = backgroundTexture[t] eePanels:ChangeBackdrop(panel) eePanels:ChangeBorderColor(panel) end,
               set = function(t) panel.texture = t eePanels:ChangeBackdrop(panel) eePanels:ChangeBorderColor(panel) end,
               --validate = val_backgroundTexture,
               order = 15,
            },
            
            backgroundBlend = {
               name = 'Background Blend', desc = 'Background Blend', type='text',
               get = function() return panel.background.gradient.blend end,
               set = function(b) panel.background.gradient.blend = blendMode[b] eePanels:ChangeTextureBlend(panel) end,
               validate = val_blendMode,
               order = 16,
            },
            
            level = {
               name = 'Level', desc = 'Panel Level', type='range',
               get = function() return panel.level end,
               set = function(l) panel.level = l eePanels:ChangeLevel(panel) end,
               min = 0, max = 20, step = 1, isPercent = false,
               order = 17,
            },
            
            strata = {
               name = 'Strata', desc = 'Panel Strata', type='text', 
               get = function() return panel.strata end,
               set = function(u) panel.strata = frameStrata[u] eePanels:ChangeStrata(panel) end,
               validate = val_frameStrata,
               order = 18,
            },
            
            width = {
               name = 'Width', desc = 'Width', type='text', usage=' ',
               get = function() return math.round(panel.width) end,
               set = function(u) panel.width = u eePanels:ChangeWidth(panel) end,
               order = 19,
            },
            
            height = {
               name = 'Height', desc = 'Height', type='text', usage=' ',
               get = function() return math.round(panel.height) end,
               set = function(u) panel.height = u eePanels:ChangeHeight(panel) end,
               order = 20,
            },
            
            x = {
               name = 'Horizontal Position', desc = 'Horizontal Position', type='text', usage=' ',
               get = function() return math.round(panel.x) end,
               set = function(u) panel.x = u eePanels:ChangePosition(panel) end,
               order = 21,
            },
            
            y = {
               name = 'Vertical Position', desc = 'Vertical Position', type='text', usage=' ',
               get = function() return math.round(panel.y) end,
               set = function(u) panel.y = u eePanels:ChangePosition(panel) end,
               order = 22,
            },
            
            space = {
               name = " ",
               type = 'header',
               order = 23,
            },
            
            remove = {
               name = 'Remove Panel', desc = 'Remove Panel', type = 'execute',
               func = function() eePanels:RemovePanel(index) end,
               order = 24,
            },
            
         },
      }
   end
end






function eePanels:Reset()
   -- Hide all existing panels
   for i in pairs(eePanels.db.profile.panels) do
      if eePanels.db.profile.panels[i].frame then 
         eePanels.db.profile.panels[i].frame:Hide() 
      end
      eePanels.db.profile.panels[i] = {}
   end
   
   -- Empty panel table
   eePanels.db.profile.panels = {}
   -- Empty panel menu table
   eePanels.menu.args.panels.args = {}
   -- Reset the default panel options
   eePanels.db.profile.defaultPanel = deepCopy(dp)
   --eePanels:ResetDB("profile")
   
   -- Force dewdrop to update
   if AceLibrary:HasInstance("Dewdrop-2.0") then
      AceLibrary:GetInstance("Dewdrop-2.0"):Refresh(1)
   end
end


function eePanels:ToggleLock()
   eePanels.db.profile.isLocked = not eePanels.db.profile.isLocked
   
   -- Unlocked: show guideFrame
   if not eePanels.db.profile.isLocked then
      for i in pairs(eePanels.db.profile.panels) do
         eePanels:CreateGuideFrame(eePanels.db.profile.panels[i],i)
      end
   -- Locked: hide guideFrame
   else
      for i in pairs(eePanels.db.profile.panels) do
         if eePanels.db.profile.panels[i].guideFrame then
            eePanels.db.profile.panels[i].guideFrame:Hide()
         end
      end
   end
   
end


function eePanels:ChangeName(panel,i)
   if not panel.name then panel.name = "" end
   if panel.guideFrame then 
      panel.guideFrame.text:SetText("Panel "..i..": "..panel.name) 
   end
end


function eePanels:ChangePosition(panel)
   panel.frame:ClearAllPoints()
   panel.frame:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", panel.x, panel.y)
end


function eePanels:ChangeWidth(panel)
   panel.frame:SetWidth(tonumber(panel.width))
   -- Fix for guideFrames not always adjust to changed panel size
   if panel.guideFrame then panel.guideFrame:SetAllPoints(panel.frame) end
end


function eePanels:ChangeHeight(panel)
   panel.frame:SetHeight(tonumber(panel.height))
   -- Fix for guideFrames not always adjust to changed panel size
   if panel.guideFrame then panel.guideFrame:SetAllPoints(panel.frame) end
end


function eePanels:ChangeBackdrop(panel)
   ChatFrame1:AddMessage("*"..panel.texture.."*")
   panel.frame:SetBackdrop({
		bgFile = panel.texture, tile = false, 
      tileSize = 32, edgeFile=panel.border.texture, edgeSize = 16,
		insets = {left = 5, right = 5, top = 5, bottom = 5},
	})
   panel.frame:SetBackdropColor(0.5,0.5,0.5,panel.textureAlpha)
end


function eePanels:CreateFrame(panel)
   panel.frame = CreateFrame("Frame", "panel", UIParent)
   panel.frame:SetFrameLevel(0)
end


function eePanels:CreateTexture(panel)
   panel.background.frame = panel.frame:CreateTexture(nil, "PARENT")
   panel.background.frame:SetPoint("TOPLEFT", panel.frame, "TOPLEFT",5,-5)
   panel.background.frame:SetPoint("BOTTOMRIGHT", panel.frame, "BOTTOMRIGHT",-5,5)
end


function eePanels:ChangeTextureBlend(panel)
   panel.background.frame:SetBlendMode(panel.background.gradient.blend)
   --panel.background.frame:SetBlendMode("ADD")
end


function eePanels:ChangeStrata(panel)
   panel.frame:SetFrameStrata(panel.strata)
end


function eePanels:ChangeLevel(panel)
   if not panel.level then panel.level = 0 end -- new property, ensure exists
   panel.frame:SetFrameLevel(panel.level)
end


function eePanels:ChangeBorderColor(panel)
   panel.frame:SetBackdropBorderColor(
      panel.border.color.r,
      panel.border.color.g,
      panel.border.color.b,
      panel.border.color.a )
end


function eePanels:ChangeBackgroundColor(panel)
   -- Display flat color
   if panel.background.style == bgColorStyle.solid then
      panel.background.frame:SetGradientAlpha(
         panel.background.gradient.orientation,
         panel.background.color.r,
         panel.background.color.g,
         panel.background.color.b,
         panel.background.color.a,
         panel.background.color.r,
         panel.background.color.g,
         panel.background.color.b,
         panel.background.color.a
      )
      panel.background.frame:SetTexture(
         panel.background.color.r,
         panel.background.color.g,
         panel.background.color.b,
         panel.background.color.a
      )
   -- Display gradient color
   else
      panel.background.frame:SetGradientAlpha(
         panel.background.gradient.orientation,
         panel.background.color.r,
         panel.background.color.g,
         panel.background.color.b,
         panel.background.color.a,
         panel.background.gradient.color.r,
         panel.background.gradient.color.g,
         panel.background.gradient.color.b,
         panel.background.gradient.color.a
      )
      panel.background.frame:SetTexture(
         panel.background.color.r,
         panel.background.color.g,
         panel.background.color.b,
         1
      )
   end
end
