--[[
-- Displays a guideFrame to move a panel by mouse
--]]
function eePanels:CreateGuideFrame(panel,i) 
   -- If the guideFrame exists, just show it
   if panel.guideFrame then
      panel.guideFrame.text:SetText( "Panel_"..i )
      panel.guideFrame:Show()
   -- Otherwise, create a new guideFrame
   else
      panel.guideFrame = CreateFrame("Frame", "position", panel.frame)
      panel.guideFrame:EnableMouse(true)
      panel.guideFrame:SetMovable(true)
      panel.guideFrame:SetFrameStrata("HIGH")
      panel.guideFrame:SetAllPoints(panel.frame)
      panel.guideFrame:SetBackdropColor(1,1,1,0)
   
      -- Create a green texture which will appear over the panel when we mouseOver
      panel.guideFrame.texture = panel.guideFrame:CreateTexture(nil, "HIGHLIGHT")
      panel.guideFrame.texture:SetAllPoints(panel.guideFrame)
      panel.guideFrame.texture:SetTexture(0,1,0)
      panel.guideFrame.texture:SetAlpha(.3)
      
      -- Display the frame number on top of our texture when we mouseOver
      panel.guideFrame.text = panel.guideFrame:CreateFontString(nil, "HIGHLIGHT")
      panel.guideFrame.text:SetFontObject(GameFontHighlightSmall)
      panel.guideFrame.text:SetPoint("CENTER", panel.guideFrame, "CENTER", 0, 0)
      panel.guideFrame.text:SetText( "Panel "..i..": "..panel.name )
      
      -- Set scripts to let us move while dragging
      panel.guideFrame:SetScript("OnMouseUp", self.StopMoving)
      panel.guideFrame:SetScript("OnMouseDown",self.StartMoving)
   
      -- Store our panel (which is the guideFrames parent) as a variable of the 
      -- guideFrame, to make it easier to figure out which panel to reposition after
      -- we've moved the guideFrame
      panel.guideFrame.parent = panel.frame
   end
end


--[[
-- Setup a panel for being moved via cursor
--]]
function eePanels:StartMoving()
   if arg1 == "LeftButton" then
      -- Attach an OnUpdate call which fires every repaint while we're moving
      this:SetScript("OnUpdate", eePanels.UpdateMove)
      -- System call
      this:StartMoving()
   end
end


--[[
-- Sets the panel to the position of it's guideFrame each time it's moved
--]]
function eePanels:UpdateMove()
   --if this.text then
      this.parent:ClearAllPoints()
      this.parent:SetPoint("TOPLEFT", UIPanel, "BOTTOMLEFT", this:GetLeft(), this:GetTop())
   --end
end


--[[
-- Stores the panel's new location in the database when a move is finished
--]]
function eePanels:StopMoving()
   -- System call
   this:StopMovingOrSizing()
   -- Remove the OnUpdate script for the guideFrame
   this:SetScript("OnUpdate", nil)
   
   -- Change panel's x,y coords in database to the guideFrame's x,y coords
   for i in pairs(eePanels.db.profile.panels) do
      if eePanels.db.profile.panels[i].guideFrame == this then
         eePanels.db.profile.panels[i].x = this:GetLeft()
         eePanels.db.profile.panels[i].y = this:GetTop()
      end
   end
end